package org.jenkinsci.plugins.p4;

import hudson.AbortException;
import hudson.FilePath.FileCallable;
import hudson.matrix.MatrixBuild;
import hudson.matrix.MatrixRun;
import hudson.model.AbstractBuild;
import hudson.model.TaskListener;
import hudson.remoting.VirtualChannel;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.jenkinsci.plugins.p4.client.CBDFileInfo;
import org.jenkinsci.plugins.p4.client.ClientHelper;
import org.jenkinsci.plugins.p4.client.ConnectionHelper;
import org.jenkinsci.plugins.p4.credentials.P4StandardCredentials;
import org.jenkinsci.plugins.p4.tagging.TagAction;
import org.jenkinsci.plugins.p4.populate.Populate;
import org.jenkinsci.plugins.p4.workspace.Workspace;

public class CheckoutTask implements FileCallable<Boolean>, Serializable {

	private static final long serialVersionUID = 1L;

	private static Logger logger = Logger.getLogger(CheckoutTask.class
			.getName());

	private final P4StandardCredentials credential;
	private final TaskListener listener;
	private final String client;

	private CheckoutStatus status;
	private int head;
	private Object buildChange;
	private int review;
	private Populate populate;
	private String cbdFileWithRevision;
	private String cbdFileHeadChange;
	private String cbdOsOption;

	/**
	 * Constructor
	 * 
	 * @param config
	 *            - Server connection details
	 * @param auth
	 *            - Server login details
	 */
	public CheckoutTask(String credentialID, Workspace config,
			TaskListener listener) {
		this.credential = ConnectionHelper.findCredential(credentialID);
		this.listener = listener;
		this.client = config.getFullName();
		this.cbdFileWithRevision = "";
		this.cbdFileHeadChange = "";
		this.cbdOsOption = config.getOSStringForCBD();
	}

	public void setBuildOpts(Workspace workspace) throws AbortException {

		ClientHelper p4 = new ClientHelper(credential, listener, client);

		try {
			// If a CBD file is specified then need to create/update the workspace required for this build
			if (!cbdFileWithRevision.equals("")) {
				if (!p4.createClient(cbdFileWithRevision, workspace)) {
					String err = "Error creating workspace: " + workspace.getFullName();
					logger.severe(err);
					listener.error(err);
					throw new AbortException(err);
				}
			}
			
			// setup the client workspace to use for the build.
			if (!p4.setClient(workspace)) {
				String err = "Undefined workspace: " + workspace.getFullName();
				logger.severe(err);
				listener.error(err);
				throw new AbortException(err);
			}

			// fetch and calculate change to sync to or review to unshelve.
			status = getStatus(workspace);
			head = p4.getClientHead();
			review = getReview(workspace);
			buildChange = getBuildChange(workspace);

		} catch (Exception e) {
			String err = "Unable to setup workspace: " + e;
			logger.severe(err);
			listener.error(err);
			throw new AbortException(err);
		} finally {
			p4.disconnect();
		}
	}

	public void setPopulateOpts(Populate populate) {
		this.populate = populate;
	}
	
	public void setBuildChange(Object change) {
		buildChange = change;
	}

	public void setCBDFileRevision(String cbdFileDepotLocation) throws AbortException {
		
		if (cbdFileDepotLocation.equals("")) {
			listener.getLogger().println("No CBD file is set to be used with this build.");
			return;
		}
		
		ClientHelper p4 = new ClientHelper(credential, listener, client);
		try {
			// Get data about the latest revision of the cbd file, and set cbdFileWithRevision and cbdFileHeadChange
			CBDFileInfo cbdInfo = p4.getCBDFileInfo(cbdFileDepotLocation);
			cbdFileWithRevision = cbdFileDepotLocation + "#" + cbdInfo.getHeadRev();
			cbdFileHeadChange = cbdInfo.getHeadChange();
			listener.getLogger().println("CBD file revision to be used is: " + cbdFileWithRevision + " (headChange of CBD file: " + cbdFileHeadChange + ")");
		} catch (Exception e) {
			String err = "Unable to set CBD file revision data: " + e;
			logger.severe(err);
			listener.error(err);
			throw new AbortException(err);
		} finally {
			p4.disconnect();
		}
	}
	
	public void setCBDFileRevisionFromParentBuild(AbstractBuild<?, ?> build) {
		MatrixBuild parentBuild = ((MatrixRun) build).getParentBuild();
		TagAction tagAction = parentBuild.getAction(TagAction.class);
		if (tagAction != null) {
			if ((tagAction.getCbdFile() != null) && (tagAction.getCbdFileHeadChange() != null)) {
				if (tagAction.getCbdFile().equals("")) {
					listener.getLogger().println("No CBD file is set to be used with this build.");
				} else {
					cbdFileWithRevision = tagAction.getCbdFile();
					cbdFileHeadChange = tagAction.getCbdFileHeadChange();
					listener.getLogger().println("Using CBD file revision as set in parent build: " + cbdFileWithRevision + " (headChange of CBD file: " + cbdFileHeadChange + ")");
				}
			}
			else {
				listener.getLogger().println("WARNING: CBD file setting from parent build is unexpectedly set to null.");
			}
		}
		else {
			listener.getLogger().println("WARNING: Failed to find parent build tagAction, so any CBD file setting will not be used.");
		}
	}
	
	/**
	 * Invoke sync on build node (master or remote node).
	 * 
	 * @return true if updated, false if no change.
	 */
	public Boolean invoke(File workspace, VirtualChannel channel)
			throws IOException {

		ClientHelper p4 = new ClientHelper(credential, listener, client);

		try {
			// test server connection
			if (!p4.isConnected()) {
				p4.log("P4: Server connection error:" + credential.getP4port());
				return false;
			}
			p4.log("Connected to server: " + credential.getP4port());

			// test client connection
			if (p4.getClient() == null) {
				p4.log("P4: Client unknown: " + client);
				return false;
			}
			p4.log("Connected to client: " + client);

			// Tidy the workspace before sync/build
			p4.tidyWorkspace(populate);

			// Sync workspace to label, head or specified change
			p4.syncFiles(buildChange, populate);

			// Unshelve review if specified
			if (status == CheckoutStatus.SHELVED) {
				p4.unshelveFiles(review);
			}
		} catch (Exception e) {
			String msg = "Unable to update workspace: " + e;
			logger.warning(msg);
			throw new AbortException(msg);
		} finally {
			p4.disconnect();
		}
		return true;
	}

	/**
	 * Get the build status for the parameter map.
	 * 
	 * @param map
	 * @return
	 */
	private CheckoutStatus getStatus(Workspace workspace) {
		CheckoutStatus status = CheckoutStatus.HEAD;
		String value = workspace.get("status");
		if (value != null && !value.isEmpty()) {
			status = CheckoutStatus.parse(value);
		}
		return status;
	}

	/**
	 * Get the sync point from the parameter map. Returns the head if no change
	 * found in the map.
	 * 
	 * @param map
	 * @return
	 */
	private Object getBuildChange(Workspace workspace) {
		// Use head as the default
		Object build = this.head;

		// if change is specified then update
		String change = workspace.get("change");
		if (change != null && !change.isEmpty()) {
			try {
				build = Integer.parseInt(change);
			} catch (NumberFormatException e) {
			}
		}

		// if label is specified then update
		String label = workspace.get("label");
		if (label != null && !label.isEmpty()) {
			try {
				// if build is a change-number passed as a label
				build = Integer.parseInt(label);
			} catch (NumberFormatException e) {
				build = label;
			}
		}

		return build;
	}

	/**
	 * Get the unshelve point from the parameter map.
	 * 
	 * @param map
	 * @return
	 */
	private int getReview(Workspace workspace) {
		int review = 0;
		String value = workspace.get("review");
		if (value != null && !value.isEmpty()) {
			try {
				review = Integer.parseInt(value);
			} catch (NumberFormatException e) {
			}
		}
		return review;
	}

	public List<Object> getChanges(Object last) {

		List<Object> changes = new ArrayList<Object>();

		// Add changes to this build.
		ClientHelper p4 = new ClientHelper(credential, listener, client);
		try {
			
			Object buildChangeToUse = buildChange;
			
			// Old CBD functionality using "p4 changes"
			/*
			if (!cbdFileHeadChange.equals("")) {
				// Using CBD workflow
				// Need the upper limit of the "changes" command to be able to indicate the revision of the CBD xml
				// file that was in use with the build (due to how the "changes" broker script is working)
				// So use the head change of the CBD xml file, if it is greater than the head of this workspace (i.e. xml
				// has changed more recently that any of the code required for the build)
				listener.getLogger().println("getChanges - buildChange set to: " + buildChange + " cbdFileHeadChange: " +  cbdFileHeadChange);
				int cbdHeadChangeInt = Integer.parseInt(cbdFileHeadChange);
				// buildChange could be a String or an int - only dealing with int here
				if (buildChange instanceof Integer) {
					int buildChangeInt = (Integer) buildChange;
					if (cbdHeadChangeInt > buildChangeInt) {
						buildChangeToUse = cbdHeadChangeInt;
						listener.getLogger().println("Changes command will use CBD file head change as upper limit.");
					}
				}
			}*/	
			
			changes = p4.listChanges(last, buildChangeToUse);
			
		} catch (Exception e) {
			String err = "Unable to get changes: " + e;
			logger.severe(err);
			listener.getLogger().println(err);
			e.printStackTrace();
		} finally {
			p4.disconnect();
		}

		// Include shelf if a review
		if (status == CheckoutStatus.SHELVED) {
			changes.add(review);
		}

		return changes;
	}

	
	public String getChangesCBD(String previousBuildCBDFileWithRevision, String previousBuildChangelistNum) {
		ClientHelper p4 = new ClientHelper(credential, listener, client);
		String changelog = "";
		
		listener.getLogger().println("CREATING CHANGE LOGS by comparing: ");
		listener.getLogger().println("Previous build " + previousBuildCBDFileWithRevision + " at " + previousBuildChangelistNum);
		listener.getLogger().println("and current build " + cbdFileWithRevision + " at " + buildChange);
		listener.getLogger().println("With os option set to " + cbdOsOption);
		
		try {
			// Test strings
			//String fromCbdFile = "//Foundry/CBD/Apps/Nuke/main_build.xml#13";
			//String toCbdFile = "//Foundry/CBD/Apps/Nuke/main_build.xml#14";
			//changelog = p4.listWsChanges(fromCbdFile, "#head", toCbdFile, "#head", cbdOsOption);
			
			String fromRev = "@" + previousBuildChangelistNum;
			String toRev = "@" + buildChange;
			changelog = p4.listWsChanges(previousBuildCBDFileWithRevision, fromRev, cbdFileWithRevision, toRev, cbdOsOption);
			
		} catch (Exception e) {
			String err = "Unable to get changes: " + e;
			logger.severe(err);
			listener.getLogger().println(err);
			e.printStackTrace();
		} finally {
			p4.disconnect();
		}
		
		return changelog;
	}
	
	public CheckoutStatus getStatus() {
		return status;
	}

	public Object getBuildChange() {
		if (status == CheckoutStatus.SHELVED) {
			return review;
		}
		return buildChange;
	}
	
	public String getCbdFileWithRevision() {
		return cbdFileWithRevision;
	}
	
	public String getCbdFileHeadChange() {
		return cbdFileHeadChange;
	}
	
	public String createSnapshotCBD() {
		// Note that failure to create a snapshot should NOT fail a build.
		ClientHelper p4 = new ClientHelper(credential, listener, client);
		String snapshot = "";
		try {
			if (buildChange instanceof Integer) {
				snapshot = p4.createSnapshot(cbdFileWithRevision, (Integer)buildChange);
			}
			else {
				listener.getLogger().println("createSnapshotCBD - buildChange is not an integer, so snapshot not created.");
			}
		} catch (Exception e) {
			String err = "Unable to create snapshot: " + e;
			logger.severe(err);
			listener.getLogger().println(err);
		} finally {
			p4.disconnect();
		}
		return snapshot;
	}
}
